package gr.cite.notification.integrationevent.inbox.tenanttouched;

import gr.cite.commons.web.oidc.principal.CurrentPrincipalResolver;
import gr.cite.notification.audit.AuditableAction;
import gr.cite.notification.common.JsonHandlingService;
import gr.cite.notification.common.scope.fake.FakeRequestScope;
import gr.cite.notification.integrationevent.inbox.EventProcessingStatus;
import gr.cite.notification.integrationevent.inbox.InboxPrincipal;
import gr.cite.notification.integrationevent.inbox.IntegrationEventProperties;
import gr.cite.notification.model.persist.TenantTouchedIntegrationEventPersist;
import gr.cite.notification.service.tenant.TenantService;
import gr.cite.tools.auditing.AuditService;
import gr.cite.tools.logging.LoggerService;
import gr.cite.tools.validation.ValidatorFactory;
import jakarta.persistence.EntityManager;
import jakarta.persistence.EntityManagerFactory;
import jakarta.persistence.EntityTransaction;
import jakarta.persistence.OptimisticLockException;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.config.ConfigurableBeanFactory;
import org.springframework.context.ApplicationContext;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import java.util.AbstractMap;
import java.util.Map;

@Component
@Scope(value = ConfigurableBeanFactory.SCOPE_PROTOTYPE)
public class TenantTouchedIntegrationEventHandlerImpl implements TenantTouchedIntegrationEventHandler {

    private static final LoggerService logger = new LoggerService(LoggerFactory.getLogger(TenantTouchedIntegrationEventHandlerImpl.class));

    protected final ApplicationContext applicationContext;

    private final JsonHandlingService jsonHandlingService;
    private final ValidatorFactory validatorFactory;

    public TenantTouchedIntegrationEventHandlerImpl(ApplicationContext applicationContext, JsonHandlingService jsonHandlingService, ValidatorFactory validatorFactory) {
        this.applicationContext = applicationContext;
        this.jsonHandlingService = jsonHandlingService;
        this.validatorFactory = validatorFactory;
    }

    @Override
    public EventProcessingStatus handle(IntegrationEventProperties properties, String message) {
        TenantTouchedIntegrationEvent event = this.jsonHandlingService.fromJsonSafe(TenantTouchedIntegrationEvent.class, message);
        if (event == null)
            return EventProcessingStatus.Error;

        TenantTouchedIntegrationEventPersist model = new TenantTouchedIntegrationEventPersist();
        model.setId(event.getId());
        model.setCode(event.getCode());
        this.validatorFactory.validator(TenantTouchedIntegrationEventPersist.TenantTouchedIntegrationEventPersistValidator.class).validateForce(model);

        EntityManager entityManager = null;
        EntityTransaction transaction = null;
        try (FakeRequestScope ignored = new FakeRequestScope()) {
            try {
                CurrentPrincipalResolver currentPrincipalResolver = this.applicationContext.getBean(CurrentPrincipalResolver.class);
                currentPrincipalResolver.push(InboxPrincipal.build(properties));

                EntityManagerFactory entityManagerFactory = this.applicationContext.getBean(EntityManagerFactory.class);
                entityManager = entityManagerFactory.createEntityManager();

                transaction = entityManager.getTransaction();
                transaction.begin();

                try {
                    TenantService tenantService = this.applicationContext.getBean(TenantService.class);
                    tenantService.persist(model, null);

                    AuditService auditService = this.applicationContext.getBean(AuditService.class);

                    auditService.track(AuditableAction.Tenant_Persist, Map.ofEntries(
                            new AbstractMap.SimpleEntry<String, Object>("model", model)
                    ));

                    transaction.commit();
                } catch (Exception e) {
                    transaction.rollback();
                    throw e;
                } finally {
                    currentPrincipalResolver.pop();
                }

                transaction.commit();
            } catch (OptimisticLockException ex) {
                // we get this if/when someone else already modified the notifications. We want to essentially ignore this, and keep working
                logger.debug("Concurrency exception getting queue outbox. Skipping: {} ", ex.getMessage());
                if (transaction != null)
                    transaction.rollback();
            } catch (Exception ex) {
                logger.error("Problem getting list of queue outbox. Skipping: {}", ex.getMessage(), ex);
                if (transaction != null)
                    transaction.rollback();
            } finally {
                if (entityManager != null)
                    entityManager.close();
            }
        } catch (Exception ex) {
            logger.error("Problem getting list of queue outbox. Skipping: {}", ex.getMessage(), ex);
        }
        return null;
    }

}
